Namespace RandomNumbers
    Friend Class RandomNumbersProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin lightweight random number generation demo " & vbCrLf)

            Dim hi As Integer = 10
            Dim lo As Integer = 0

            Console.WriteLine("Generating 1000 ints in [0, 9] using Lehmer RNG ")
            Dim counts(9) As Integer
            Dim lehmer As New LehmerRng(1)
            For i = 0 To 999
                Dim x As Double = lehmer.Next
                Dim ri As Integer = CInt(Fix((hi - lo) * x + lo))
                counts(ri) += 1
            Next i
            Console.Write("Counts:  ")
            For i = 0 To 9
                Console.Write(counts(i) & " ")
            Next i
            Console.WriteLine(vbCrLf & vbCrLf)

            Console.WriteLine("Generating 1000 ints in [0, 9] using Wichmann-Hill RNG ")
            counts = New Integer(9) {}
            Dim wich As New WichmannRng(1)
            For i = 0 To 999
                Dim x As Double = wich.Next
                Dim ri As Integer = CInt(Fix((hi - lo) * x + lo))
                counts(ri) += 1
            Next i
            Console.Write("Counts:  ")
            For i = 0 To 9
                Console.Write(counts(i) & " ")
            Next i
            Console.WriteLine(vbCrLf & vbCrLf)

            Console.WriteLine("Generating 1000 ints in [0, 9] using Linear Congruential RNG ")
            counts = New Integer(9) {}
            Dim lincon As New LinearConRng(0)
            For i = 0 To 999
                Dim x As Double = lincon.Next
                Dim ri As Integer = CInt(Fix((hi - lo) * x + lo))
                counts(ri) += 1
            Next i
            Console.Write("Counts:  ")
            For i = 0 To 9
                Console.Write(counts(i) & " ")
            Next i
            Console.WriteLine(vbCrLf & vbCrLf)

            Console.WriteLine("Generating 1000 ints in [0, 9] using Lagged Fibonacci RNG ")
            counts = New Integer(9) {}
            Dim lagged As New LaggedFibRng(0)
            For i = 0 To 999
                Dim x As Double = lagged.Next
                Dim ri As Integer = CInt(Fix((hi - lo) * x + lo))
                counts(ri) += 1
            Next i
            Console.Write("Counts:  ")
            For i = 0 To 9
                Console.Write(counts(i) & " ")
            Next i
            Console.WriteLine(vbCrLf & vbCrLf)

            Console.WriteLine("End demo " & vbCrLf)
            Console.ReadLine()
        End Sub ' Main
    End Class ' class Program

    Public Class LehmerRng
        Private Const a As Integer = 16807
        Private Const m As Integer = 2147483647
        Private Const q As Integer = 127773
        Private Const r As Integer = 2836
        Private seed As Integer

        Public Sub New(ByVal seed As Integer)
            If seed <= 0 OrElse seed = Integer.MaxValue Then
                Throw New Exception("Bad seed")
            End If
            Me.seed = seed
        End Sub

        Public Function [Next]() As Double
            Dim hi As Integer = seed \ q
            Dim lo As Integer = seed Mod q
            seed = (a * lo) - (r * hi)
            If seed <= 0 Then
                seed = seed + m
            End If
            Return (seed * 1.0) / m
        End Function
    End Class ' class LehmerRng

    Public Class WichmannRng
        Private s1, s2, s3 As Integer

        Public Sub New(ByVal seed As Integer)
            If seed <= 0 OrElse seed > 30000 Then
                Throw New Exception("Bad seed")
            End If
            s1 = seed
            s2 = seed + 1
            s3 = seed + 2
        End Sub

        Public Function [Next]() As Double

            s1 = 171 * (s1 Mod 177) - 2 * (s1 \ 177)
            If s1 < 0 Then
                s1 += 30269
            End If

            s2 = 172 * (s2 Mod 176) - 35 * (s2 \ 176)
            If s2 < 0 Then
                s2 += 30307
            End If

            s3 = 170 * (s3 Mod 178) - 63 * (s3 \ 178)
            If s3 < 0 Then
                s3 += 30323
            End If

            Dim r As Double = (s1 * 1.0) / 30269 + (s2 * 1.0) / 30307 + (s3 * 1.0) / 30323
            Return r - Math.Truncate(r) ' orig uses % 1.0
        End Function
    End Class ' class WichmannRng

    Public Class LinearConRng
        Private Const a As Long = 25214903917
        Private Const c As Long = 11
        ' m = 2^48
        Private seed As Long

        Public Sub New(ByVal seed As Long)
            If seed < 0 Then
                Throw New Exception("Bad seed")
            End If
            Me.seed = seed
        End Sub

        Private Function next_Renamed(ByVal bits As Integer) As Integer
            seed = (seed * a + c) And ((1L << 48) - 1) Mod &HFF
            Return CInt(Fix(seed >> (48 - bits))) ' high order bits set to sign bit (0)
        End Function

        Public Function [Next]() As Double
            Return ((CLng(Fix(next_Renamed(26))) << 27) + next_Renamed(27)) / CDbl(1L << 53)
        End Function
    End Class ' class LinearConRng

    Public Class LaggedFibRng
        Private Const k As Integer = 10 ' largest magintude"-index"
        Private Const j As Integer = 7 ' other "-index"
        Private Const m As Integer = 2147483647 ' 2^31 - 1 = maxint

        Private vals As List(Of Integer) = Nothing
        Private seed As Integer

        Public Sub New(ByVal seed As Integer)
            vals = New List(Of Integer)
            For i  = 0 To k
                vals.Add(seed)
            Next i
            If seed Mod 2 = 0 Then ' at least one seed must be odd
                vals(0) = 11
            End If

            ' burn some values away
            For ct  = 0 To 999
                Dim dummy As Double = Me.Next
            Next ct
        End Sub ' ctor

        Public Function [Next]() As Double
            Dim left As Integer = vals(0) Mod m ' x-big
            Dim right As Integer = vals(k - j) Mod m ' x-other
            Dim sum As Long = CLng(Fix(left)) + CLng(Fix(right)) ' prevent overflow

            seed = CInt(Fix(sum Mod m)) ' because m is int, result has int range
            vals.Insert(k + 1, seed) ' add new val at end
            vals.RemoveAt(0) ' delete now irrelevant 0 val
            Return (1.0 * seed) / m
        End Function
    End Class ' LaggedFibRng
End Namespace ' ns
